from skimage import io, color
from PIL import Image

import numpy as np
import cv2

from scipy.interpolate import pchip_interpolate
from scipy.optimize import minimize
from scipy.optimize import Bounds

try:
    from utils import *
except ImportError:
    from .utils import *

import warnings
warnings.filterwarnings("ignore", category=RuntimeWarning)

def color_optimization( palette, weights, target_color, palette_cons, verbose=False ):
    '''
    Given:
        `palette`: A #p-by-2 array
        `weights`: A array of shape (C, #p)
        `target_colors`: A array of shape (C, 2)
        `palette_cons`: A list of palette constraints. Like: [ [j1, (a1,b1)], [j2, (a2,b2)], ... ]
        `verbose`: toggle for displaying optimization results
    
    Return:
        `palette_opt`: A #p-by-2 array from optimization results
    '''
    np.set_printoptions( suppress = True )
    if np.size( weights ) != 0:
        assert palette.shape[0] == weights.shape[1]
        assert palette.shape[1] == target_color.shape[1]
        assert weights.shape[0] == target_color.shape[0]
    else:
        weights = np.zeros( (1,palette.shape[0]) )
        target_color = np.zeros( (1,2) )
        
    # set up palette constraints
    pc = np.zeros( (len(palette_cons), 2) ) 
    ind = np.zeros( len(palette_cons) ).astype( int )
    for i, (p_indx, cons) in enumerate( palette_cons ): 
        pc[i, :] = cons
        ind[i] = p_indx
    
    # penalize grey on sparsity energy to avoid optimizer prefering to choose grey 
    penal_weights = np.eye( palette.shape[0] )
    penal_weights[-1, -1] = 3


    ### objective function
    def func( vec_delp ):
        delp = vec_delp.reshape( palette.shape )
        delp_penal = penal_weights @ delp
        return np.sum( np.linalg.norm( delp_penal, axis = 1 ) )     # optimize for palette's sparsity
    
    ### image-space constraint
    def color_constraints( vec_delp ):
        delp = vec_delp.reshape( palette.shape )
        rec_colors = weights @ ( palette + delp )
        return 1 - np.linalg.norm( rec_colors - target_color, axis = 1 )
    
    ### constraints on palette color
    def palette_constraints( vec_delp ):
        delp = vec_delp.reshape( palette.shape )
        new_palette = palette + delp
        return np.linalg.norm( new_palette[ ind ] - pc, axis = 1 )
    
    bounds = Bounds( -128 - palette.flatten(), 127 - palette.flatten() )
    
    if len( palette_cons ) == 0:
        cons = [ {'type': 'ineq', 'fun': color_constraints} ]
    else:
        # use inequality on color constraints to avoid singularity
        # (in practice/numerically, it's ok since our sparsity energy is going to dominate this.)
        cons = [ {'type': 'ineq', 'fun': color_constraints}, {'type': 'eq', 'fun': palette_constraints} ]
    
    x0 = np.zeros( palette.shape[0] * palette.shape[1] )
    res = minimize( func, x0, method='SLSQP', constraints=cons, options={'ftol': 1e-6, 'disp': False, 'maxiter': 300}, bounds=bounds )
    print( res )
    delp = ( res.x ).reshape( palette.shape )
    
    if verbose:
        print( '-- Original weights:\n', weights )
        print( '-- Original color:\n', weights @ palette )
        print( '-- Target color:\n', target_color )
        print( '-- Reconstructed color:\n', weights @ ( delp + palette ) )
        
    return delp + palette





def main():
    import argparse
    parser = argparse.ArgumentParser( description = 'optimization on entire sparsity.' )
    parser.add_argument( 'img', help = 'The path to the input image.' )
    args = parser.parse_args()
    
    img = color.rgb2lab( np.array( Image.open( args.img ) ) / 255. )   
    L0 = img[:, :, 0] / 100
    ab = img[:, :, 1:]
    
    ab_palette = extract_AB_palette( img, 4 )
    weights = extract_weights( ab_palette, img ).reshape( (img.shape[0], img.shape[1], 5) )
    
    i, j = 100, 80
    
    constraints_weight = np.array( [[0.36522842, 0.,         0.,         0.16060654, 0.47416504]] )
    constraints_weight = np.array( [] )
    color_targets = np.array( [[ 7.93090111, 25.98326811]] )
    color_targets = np.array([])
    palette_target = [[1, np.array([  3.59220216, -57.98019723])]]
    palette_target = []
    palette_opt = color_optimization( ab_palette, constraints_weight, color_targets, palette_target )
    
    
if __name__ == '__main__':
    main()