from skimage import io, color
from PIL import Image

import numpy as np
import cv2

from scipy.interpolate import pchip_interpolate
from scipy.optimize import minimize
from scipy.optimize import Bounds

try:
    from utils import *
except ImportError:
    from .utils import *

import warnings
warnings.filterwarnings("ignore", category=RuntimeWarning)

def color_optimization( palette, weights, target_color, verbose=False ):
    '''
    Given:
        `palette`: A #p-by-2 array
        `weights`: A array of shape (C, #p)
        `target_colors`: A array of shape (C, 2)
        `verbose`: toggle for displaying optimization results
    
    Return:
        `palette_opt`: A #p-by-2 array from optimization results
    '''
    
    assert palette.shape[0] == weights.shape[1]
    assert palette.shape[1] == target_color.shape[1]
    assert weights.shape[0] == target_color.shape[0]
    
    ### objective function
    def func( vec_delp ):
        delp = vec_delp.reshape( palette.shape )
        rec_colors = weights @ ( palette + delp )
        return sum( np.linalg.norm( delp, axis = 1 ) )     # optimize for palette's sparsity
    
    def color_constraints( vec_delp ):
        delp = vec_delp.reshape( palette.shape )
        rec_colors = weights @ ( palette + delp )
        return np.linalg.norm( rec_colors - target_color, axis = 1 )
    
    
    bounds = Bounds( -128 - palette.flatten(), 127 - palette.flatten() )
    cons = [ {'type': 'eq', 'fun': color_constraints} ]
    
    x0 = np.zeros( palette.shape[0] * palette.shape[1] )
    res = minimize( func, x0, method='SLSQP', constraints=cons, options={'ftol': 1e-6, 'disp': False, 'maxiter': 300}, bounds=bounds )
    delp = ( res.x ).reshape( palette.shape )
    
    if verbose:
        print( '-- Original weights:', weights )
        print( '-- Original color: ', weights @ palette )
        print( '-- Target color: ', target_color )
        print( '-- Reconstructed color: ', weights @ ( delp + palette ) )
        
    return delp + palette

def lum_optimization( L0, L_target, control_palette, weights, verbose=False ):
    '''
    Given:
        `L0`: A C-by-1 array of original luminances (range: [0, 1])
        `L_target`: A C-by-1 array of target luminances corresponding to L0 (range: [0, 1])
        `control_palette`: A #P-by-C-2 array of controls points for each palette
        `weights`: A array of shape (C, #p)
        `verbose`: toggle for displaying optimization results
    
    Return:
        `new_control_palette`: A #P-by-C-2 array of new `control_palette`
    '''
    np.set_printoptions( suppress = True )
    assert len( control_palette.shape ) == 3
    assert L0.shape[0] == L_target.shape[0] == weights.shape[0]
    
    num_cons = L0.shape[0]
    num_func = control_palette.shape[0]
    
    # initial function values at L0 for each shape function
    discrete_vals = np.zeros( (num_cons, num_func) )
    for c in range( num_cons ):
        for i in range( num_func ):
            discrete_vals[c][i] = pchip_interpolate( control_palette[i][:,0], control_palette[i][:,1], L0[c] )
    
    # initialization for vertical changes for each shape function
    delta_function = np.zeros_like( discrete_vals ).flatten()
    
    def loss( delta_func ):
        # L2,1-norm
        delta_func = delta_func.reshape( discrete_vals.shape )
        return np.sum( np.linalg.norm( delta_func, axis = 1 ) )
    
    def lum_constraints( delta_func ):
        # Σ_i w_i (f_i + Δf_i)(L0) = L'
        delta_func = delta_func.reshape( discrete_vals.shape )
        return np.sum( weights * ( discrete_vals + delta_func ), axis = 1 ) - L_target
    
    # -f_i(L0) <= Δf_i(L0) <= 1 - f_i(L0)
    bounds = Bounds( -discrete_vals.flatten(), 1-discrete_vals.flatten() )
    cons = [ {'type': 'eq', 'fun': lum_constraints} ]
    
    res = minimize( loss, delta_function, method='SLSQP', constraints=cons, options={'ftol': 1e-6, 'disp': False, 'maxiter': 300}, bounds=bounds )
    delta_function = ( res.x ).reshape( discrete_vals.shape )
    
    if verbose:
        print( '-- Original weights:', weights )
        print( '-- Original luminance: ', L0 )
        print( '-- Target color: ', L_target )
        print( '-- Reconstructed error: ', lum_constraints( delta_function ) )
    
    # update control points palette
    new_control_pts_x = np.tile( L0, num_func )
    new_control_pts_y = np.ravel( delta_function + discrete_vals, order = 'F' )
    new_control_pts = np.vstack( (new_control_pts_x, new_control_pts_y) ).T.reshape( (num_func, num_cons, 2) )
    
    # create new control palette
    new_control_palette = np.zeros( (num_func, num_cons+2, 2) )
    new_control_palette[:, -1, :] = 1
    new_control_palette[:, 1:-1, :] = new_control_pts
    
    return new_control_palette

def reconstruct_luminance_from_control_points( L0, control_palette, weights ):
    '''
    Given:
        `L0`: A N-by-M array of original luminance
        `control_palette`: A #P-by-C-2 array of controls points for each palette
        `weights`: A N-by-M-by-5 array of weights
    
    Return:
        Reconstructed Luminance
    '''
    L = L0.reshape( -1 )
    w = weights.reshape( -1, 5 )
    L_ = np.zeros_like( w )
    
    for i in range( L_.shape[1] ):
        ind = control_palette[i][:,0].argsort()
        x = control_palette[i][:,0][ind]
        y = control_palette[i][:,1][ind]
        L_[:, i] = pchip_interpolate( x, y, L )
    return np.sum( w * L_, axis = 1 ).reshape( L0.shape )


def function_plot_for_constraint( L_cons, control_palette, w_c ):
    '''
    Note: This function is only able to plot function behaviors according to one constraint.
          It might not make sense to plot a function for multiple constraints at the same time,
          since different constraints have different weights.

    Given:
        `L_cons`: A C-by-2 array of constraints
        `control_palette`: A #P-by-C-2 array of controls points for each palette
        `w_c`: A array of shape (C, #p)
    
    Return:
        No returns. Just plot according to a specific constraint.
    '''
    import matplotlib.pyplot as plt
    
    indx = 0
    w = w_c[ indx ]
    L = np.linspace( 0, 1, 500 )
    
    fig, ax = plt.subplots( nrows = 2, ncols = 3 )
    fig.tight_layout( pad = 1.0 )
    axes = ax.flatten()
    
    all_mapped_L = 0
    for i in range( control_palette.shape[0]+1 ):
        if i == 5:
            axes[i].set_title( 'Entire Function', fontdict = {'fontsize': 7} )
            axes[i].plot( L, all_mapped_L, label = "pchip interpolation" )
        else:
            new_fi = pchip_interpolate( control_palette[i][:,0], control_palette[i][:,1], L )
            all_mapped_L += w[i] * new_fi
            
            axes[i].set_title( 'Palette ' + str(i), fontdict = {'fontsize': 7} )
            axes[i].plot( L, new_fi, label = "pchip interpolation" )
            
        axes[i].axvline( x = L_cons[indx][0], linewidth = 1.5, color = 'r', label = 'L0', linestyle=':' )
        axes[i].axhline( y = L_cons[indx][1], linewidth = 1.5, color = 'g', label = 'L0\'', linestyle=':' )
        axes[i].legend( loc="upper right", prop={'size': 5} )
    plt.show()
    
    
def main():
    import argparse
    parser = argparse.ArgumentParser( description = 'shadow/highlight testing.' )
    parser.add_argument( 'img', help = 'The path to the input image.' )
    args = parser.parse_args()
    
    img = color.rgb2lab( np.array( Image.open( args.img ) ) / 255. )   
    ab_palette = extract_AB_palette( img )
    weights = extract_weights( ab_palette, img ).reshape( (img.shape[0], img.shape[1], 5) )
    
    L0 = img[:, :, 0] / 100
    ab = weights @ ab_palette
    
    i, j = 100, 52 
    p_weights = np.array( [weights[i, j]] )
    
    target_color = np.array( [ab[i, j]] )
    target_lum = 0.1
    
    print( 'Original luminance: ', L0[i, j] )
    print( 'Target luminance: ', target_lum )
    
    print( 'Original color: ', ab[i, j] )
    print( 'Target color: ', target_color )
    
    ###-------
    
    # hard-code testing
    #p_weights = np.array([[0.32486977, 0., 0., 0.09290534, 0.58222489]])
    #target_color = np.array([[ 76.01346064, -74.16865215]])
    
    # optimize color palette
    palette_opt = color_optimization( ab_palette, p_weights, target_color )
    
    # tiles up fixed points for each shape function ( we need to pin down (0,0) and (1,1) )
    control_palette = np.tile( np.array( [[0,0], [1,1]] ), (5, 1, 1) )
    
    L_constraints = np.array( [[L0[i,j], target_lum]] ) # An C-by-2 array of constraints
    control_palette = lum_optimization( L_constraints[:, 0], L_constraints[:, 1], control_palette, p_weights )
    
    
    # reconstruct back luminance
    recon_lum = get_recon_lum( control_palette, weights.reshape(-1,5), L0 ) * 100
    recon_img = get_recon_img( palette_opt, weights, recon_lum )
    print( 'Final reconstructed lum: ', recon_lum[i,j] / 100 )
    
    ### save result
    recon_img = ( color.lab2rgb( recon_img ) * 255. ).clip( 0, 255 ).astype( np.uint8 )
    Image.fromarray( recon_img ).save( '../test-edit.png' )
    
    function_plot_for_constraint( L_constraints, control_palette, p_weights )

if __name__ == '__main__':
    main()